from __future__ import absolute_import
from __future__ import division

import copy

from .otpkpolicy import OTPKPolicy

class DefaultOTPKPolicy(OTPKPolicy):
    """
    An implementation of the OTPKPolicy with a default ruleset that slightly prefers
    usability over security.

    These are the rules:
    * Never delete an OTPK because of messages that came from some sort of storage
        mechanism like MAM
    * Never delete an OTPK as long as you have not sent a single answer
    * Only delete an OTPK if at least two answers were sent with a delay of at least 24
        hours between them

    With this ruleset possible attackers are prevented from permanently reusing an OTPK,
    while real-world use-cases should never result in lost messages because of deleted
    OTPKs.

    You can use the additional_information parameter of the SessionManager.decryptMessage
    method to declare whether a message came from some storage mechanism like MAM or not.
    To do so, pass additional_information like this::

        additional_information = {
            "from_storage": boolean
        }
    """

    @staticmethod
    def decideOTPK(preKeyMessages):
        pkms = copy.deepcopy(preKeyMessages)

        # Normalize the additional_information to contain an empty dict instead of None
        for pkm in pkms:
            if pkm["additional_information"] == None:
                pkm["additional_information"] = {}

        # Normalize the from_storage information
        for pkm in pkms:
            if not "from_storage" in pkm["additional_information"]:
                pkm["additional_information"]["from_storage"] = False

        # Filter out messages that were retreived from storage mechanisms
        pkms = list(filter(
            lambda pkm: not pkm["additional_information"]["from_storage"],
            pkms
        ))

        # Collect all answers
        answers = []

        for pkm in pkms:
            answers += pkm["answers"]

        # Check whether at least two answers were sent
        if len(answers) < 2:
            return True

        # Check whether at least 24 hours passed between the two answers
        elapsed_seconds = max(answers) - min(answers)
        elapsed_minutes = elapsed_seconds / 60
        elapsed_hours   = elapsed_minutes / 60

        if elapsed_hours < 24:
            return True

        # Otherwise, all conditions are met to delete the OTPK
        return False



from __future__ import absolute_import

import base64

def make(backend):
    class ExtendedDoubleRatchet(backend.DoubleRatchet):
        def __init__(self, other_ik, *args, **kwargs):
            super(ExtendedDoubleRatchet, self).__init__(*args, **kwargs)

            self.__other_ik = other_ik

        def serialize(self):
            return {
                "super"    : super(ExtendedDoubleRatchet, self).serialize(),
                "other_ik" : base64.b64encode(self.__other_ik).decode("US-ASCII")
            }

        @classmethod
        def fromSerialized(cls, serialized, *args, **kwargs):
            self = super(ExtendedDoubleRatchet, cls).fromSerialized(
                serialized["super"],
                *args,
                **kwargs
            )

            self.__other_ik = base64.b64decode(serialized["other_ik"].encode("US-ASCII"))

            return self

        @property
        def ik(self):
            return self.__other_ik

    return ExtendedDoubleRatchet



from __future__ import absolute_import
from __future__ import division

import base64
import copy
import logging
import os
import sys
import time

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from . import promise
from . import storagewrapper
from .exceptions import *
from .x3dhdoubleratchet import make as make_X3DHDoubleRatchet

import x3dh

# This makes me sad
if sys.version_info[0] == 3:
    string_type = str
else:
    string_type = basestring

def d(*args, **kwargs):
    logging.getLogger("omemo.SessionManager").debug(*args, **kwargs)

def w(*args, **kwargs):
    logging.getLogger("omemo.SessionManager").warning(*args, **kwargs)

def e(*args, **kwargs):
    logging.getLogger("omemo.SessionManager").error(*args, **kwargs)

def checkSelf(self, *args, **kwargs):
    return self._storage.is_async

def checkPositionalArgument(position):
    def _checkPositionalArgument(*args, **kwargs):
        return args[position].is_async

    return _checkPositionalArgument

class SessionManager(object):
    ################################
    # construction and preparation #
    ################################

    @classmethod
    @promise.maybe_coroutine(checkPositionalArgument(1))
    def create(
        cls,
        storage,
        otpk_policy,
        backend,
        my_bare_jid,
        my_device_id,
        inactive_per_jid_max = 15,
        inactive_global_max  =  0,
        inactive_max_age     =  0
    ):
        self = cls()

        # Store the parameters
        self._storage = storagewrapper.StorageWrapper(storage)

        self.__otpk_policy = otpk_policy

        self.__backend = backend
        self.__X3DHDoubleRatchet = make_X3DHDoubleRatchet(self.__backend)

        self.__my_bare_jid  = my_bare_jid
        self.__my_device_id = my_device_id

        self.__inactive_per_jid_max = inactive_per_jid_max
        self.__inactive_global_max  = inactive_global_max
        self.__inactive_max_age     = inactive_max_age

        # Prepare the caches
        self.__state = None

        self.__sessions_cache = {}
        self.__devices_cache  = {}
        self.__trust_cache    = {}

        yield self.__prepare()

        promise.returnValue(self)

    @promise.maybe_coroutine(checkSelf)
    def __prepare(self):
        state = yield self._storage.loadState()
        if state == None:
            self.__state = self.__X3DHDoubleRatchet()

            yield self._storage.storeState(self.__state.serialize())

            yield self.__storeActiveDevices(self.__my_bare_jid, [ self.__my_device_id ])
        else:
            self.__state = self.__X3DHDoubleRatchet.fromSerialized(state)

        own_data = yield self._storage.loadOwnData()
        if own_data == None:
            yield self._storage.storeOwnData(self.__my_bare_jid, self.__my_device_id)
        else:
            if (not self.__my_bare_jid  == own_data["own_bare_jid"] or
                not self.__my_device_id == own_data["own_device_id"]):
                raise InconsistentInfoException(
                    "Given storage is only usable for jid \"" + own_data["own_bare_jid"] +
                    "\" on device " + str(own_data["own_device_id"]) + "."
                )

    ##############
    # encryption #
    ##############

    @promise.maybe_coroutine(checkSelf)
    def encryptMessage(
        self,
        bare_jids,
        plaintext,
        bundles = None,
        expect_problems = None
    ):
        # Dirty hack to access ciphertext from _encryptMessage
        ciphertext = []

        def _encryptMessage(aes_gcm):
            ciphertext.append(aes_gcm.update(plaintext) + aes_gcm.finalize())

        encrypted = yield self.encryptKeyTransportMessage(
            bare_jids,
            _encryptMessage,
            bundles,
            expect_problems
        )

        encrypted["payload"] = ciphertext[0]

        promise.returnValue(encrypted)

    @promise.maybe_coroutine(checkSelf)
    def encryptRatchetForwardingMessage(
        self,
        bare_jids,
        bundles = None,
        expect_problems = None
    ):
        encrypted = yield self.encryptKeyTransportMessage(
            bare_jids,
            lambda aes_gcm: aes_gcm.finalize(),
            bundles,
            expect_problems
        )

        promise.returnValue(encrypted)

    @promise.maybe_coroutine(checkSelf)
    def encryptKeyTransportMessage(
        self,
        bare_jids,
        encryption_callback,
        bundles = None,
        expect_problems = None
    ):
        """
        bare_jids: iterable<string>
        encryption_callback: A function which is called using an instance of cryptography.hazmat.primitives.ciphers.CipherContext, which you can use to encrypt any sort of data. You don't have to return anything.
        bundles: { [bare_jid: string] => { [device_id: int] => ExtendedPublicBundle } }
        expect_problems: { [bare_jid: string] => iterable<int> }

        returns: {
            iv: bytes,
            sid: int,
            keys: {
                [bare_jid: string] => {
                    [device: int] => {
                        "data"    : bytes,
                        "pre_key" : boolean
                    }
                }
            }
        }
        """

        yield self.runInactiveDeviceCleanup()

        #########################
        # parameter preparation #
        #########################

        if isinstance(bare_jids, string_type):
            bare_jids = set([ bare_jids ])
        else:
            bare_jids = set(bare_jids)

        if bundles == None:
            bundles = {}

        if expect_problems == None:
            expect_problems = {}
        else:
            for bare_jid in expect_problems:
                expect_problems[bare_jid] = set(expect_problems[bare_jid])

        # Add the own bare jid to the set of jids
        bare_jids.add(self.__my_bare_jid)

        ########################################################
        # check all preconditions and prepare missing sessions #
        ########################################################

        problems = []

        # Prepare the lists of devices to encrypt for
        encrypt_for = {}

        for bare_jid in bare_jids:
            devices = yield self.__loadActiveDevices(bare_jid)

            if len(devices) == 0:
                problems.append(NoDevicesException(bare_jid))
            else:
                encrypt_for[bare_jid] = devices

        # Remove the sending devices from the list
        encrypt_for[self.__my_bare_jid].remove(self.__my_device_id)

        # Check whether all required bundles are available
        for bare_jid, devices in encrypt_for.items():
            missing_bundles = set()

            for device in devices:
                session = yield self.__loadSession(bare_jid, device)

                if session == None:
                    if not device in bundles.get(bare_jid, {}):
                        missing_bundles.add(device)

            devices -= missing_bundles

            for device in missing_bundles:
                if not device in expect_problems.get(bare_jid, set()):
                    problems.append(MissingBundleException(bare_jid, device))

        # Check for missing sessions and simulate the key exchange
        for bare_jid, devices in encrypt_for.items():
            key_exchange_problems = {}

            for device in devices:
                # Load the session
                session = yield self.__loadSession(bare_jid, device)

                # If no session exists, create a new session
                if session == None:
                    # Get the required bundle
                    bundle = bundles[bare_jid][device]

                    try:
                        # Build the session, discarding the result afterwards. This is
                        # just to check that the key exchange works.
                        self.__state.getSharedSecretActive(bundle)
                    except x3dh.exceptions.KeyExchangeException as e:
                        key_exchange_problems[device] = str(e)

            encrypt_for[bare_jid] -= set(key_exchange_problems.keys())

            for device, message in key_exchange_problems.items():
                if not device in expect_problems.get(bare_jid, set()):
                    problems.append(KeyExchangeException(
                        bare_jid,
                        device,
                        message
                    ))

        # Check the trust for each device
        for bare_jid, devices in encrypt_for.items():
            untrusted = []

            for device in devices:
                # Load the session
                session = yield self.__loadSession(bare_jid, device)

                # Get the identity key of the recipient
                other_ik = bundles[bare_jid][device].ik if session == None else session.ik

                if not (yield self.__checkTrust(bare_jid, device, other_ik)):
                    untrusted.append((device, other_ik))

            devices -= set(map(lambda x: x[0], untrusted))

            for device, other_ik in untrusted:
                if not device in expect_problems.get(bare_jid, set()):
                    problems.append(UntrustedException(bare_jid, device, other_ik))

        # Check for jids with no eligible devices
        for bare_jid, devices in list(encrypt_for.items()):
            # Skip this check for my own bare jid
            if bare_jid == self.__my_bare_jid:
                continue

            if len(devices) == 0:
                problems.append(NoEligibleDevicesException(bare_jid))
                del encrypt_for[bare_jid]

        # If there were and problems, raise an Exception with a list of those.
        if len(problems) > 0:
            raise EncryptionProblemsException(problems)

        ##############
        # encryption #
        ##############

        # Prepare AES-GCM key and IV
        aes_gcm_iv  = os.urandom(16)
        aes_gcm_key = os.urandom(16)

        # Create the AES-GCM instance
        aes_gcm = Cipher(
            algorithms.AES(aes_gcm_key),
            modes.GCM(aes_gcm_iv),
            backend=default_backend()
        ).encryptor()

        # Encrypt the plain data
        encryption_callback(aes_gcm)

        # Store the tag
        aes_gcm_tag = aes_gcm.tag

        # {
        #     [bare_jid: string] => {
        #         [device: int] => {
        #             "data"    : bytes,
        #             "pre_key" : boolean
        #         }
        #     }
        # }
        encrypted_keys = {}

        for bare_jid, devices in encrypt_for.items():
            encrypted_keys[bare_jid] = {}

            for device in devices:
                # Note whether this is a response to a PreKeyMessage
                if self.__state.hasBoundOTPK(bare_jid, device):
                    self.__state.respondedTo(bare_jid, device)
                    yield self._storage.storeState(self.__state.serialize())

                # Load the session
                session = yield self.__loadSession(bare_jid, device)

                # If no session exists, this will be a PreKeyMessage
                pre_key = session == None

                # Create a new session                
                if pre_key:
                    # Get the required bundle
                    bundle = bundles[bare_jid][device]

                    # Build the session
                    session_and_init_data = self.__state.getSharedSecretActive(bundle)
                    
                    session = session_and_init_data["dr"]
                    session_init_data = session_and_init_data["to_other"]

                # Encrypt the AES GCM key and tag
                encrypted_data = session.encryptMessage(aes_gcm_key + aes_gcm_tag)

                # Store the new/changed session
                yield self.__storeSession(bare_jid, device, session)

                # Serialize the data into a simple message format
                serialized = self.__backend.WireFormat.messageToWire(
                    encrypted_data["ciphertext"],
                    encrypted_data["header"],
                    { "DoubleRatchet": encrypted_data["additional"] }
                )

                # If it is a PreKeyMessage, apply an additional step to the serialization.
                if pre_key:
                    serialized = self.__backend.WireFormat.preKeyMessageToWire(
                        session_init_data,
                        serialized,
                        { "DoubleRatchet": encrypted_data["additional"] }
                    )

                # Add the final encrypted and serialized data.
                encrypted_keys[bare_jid][device] = {
                    "data"    : serialized,
                    "pre_key" : pre_key
                }

        promise.returnValue({
            "iv"   : aes_gcm_iv,
            "sid"  : self.__my_device_id,
            "keys" : encrypted_keys
        })

    ##############
    # decryption #
    ##############

    @promise.maybe_coroutine(checkSelf)
    def decryptMessage(
        self,
        bare_jid,
        device,
        iv,
        message,
        is_pre_key_message,
        ciphertext,
        additional_information = None,
        allow_untrusted = False
    ):
        aes_gcm = yield self.decryptKeyTransportMessage(
            bare_jid,
            device,
            iv,
            message,
            is_pre_key_message,
            additional_information,
            allow_untrusted
        )

        promise.returnValue(aes_gcm.update(ciphertext) + aes_gcm.finalize())

    @promise.maybe_coroutine(checkSelf)
    def decryptRatchetForwardingMessage(
        self,
        bare_jid,
        device,
        iv,
        message,
        is_pre_key_message,
        additional_information = None,
        allow_untrusted = False
    ):
        aes_gcm = yield self.decryptKeyTransportMessage(
            bare_jid,
            device,
            iv,
            message,
            is_pre_key_message,
            additional_information,
            allow_untrusted
        )

        aes_gcm.finalize()

    @promise.maybe_coroutine(checkSelf)
    def decryptKeyTransportMessage(
        self,
        bare_jid,
        device,
        iv,
        message,
        is_pre_key_message,
        additional_information = None,
        allow_untrusted = False
    ):
        yield self.runInactiveDeviceCleanup()

        if is_pre_key_message:
            # Unpack the pre key message data
            message_and_init_data = self.__backend.WireFormat.preKeyMessageFromWire(
                message
            )

            other_ik = message_and_init_data["session_init_data"]["ik"]

            # Before doing anything else, check the trust
            if not allow_untrusted:
                if not (yield self.__checkTrust(bare_jid, device, other_ik)):
                    raise UntrustedException(bare_jid, device, other_ik)

            # Prepare the DoubleRatchet
            try:
                session = self.__state.getSharedSecretPassive(
                    message_and_init_data["session_init_data"],
                    bare_jid,
                    device,
                    self.__otpk_policy,
                    additional_information
                )
            except x3dh.exceptions.KeyExchangeException as e:
                raise KeyExchangeException(bare_jid, device, str(e))

            # Store the changed state
            yield self._storage.storeState(self.__state.serialize())

            # Store the new session
            yield self.__storeSession(bare_jid, device, session)

            # Unpack the "normal" message that was wrapped into the PreKeyMessage
            message = message_and_init_data["message"]

        # Load the session
        session = yield self.__loadSession(bare_jid, device)
        if session == None:
            raise NoSessionException(bare_jid, device)

        # Before doing anything else, check the trust
        if not allow_untrusted:
            if not (yield self.__checkTrust(bare_jid, device, session.ik)):
                raise UntrustedException(bare_jid, device, session.ik)

        # Now that the trust was checked, go on with normal processing
        if not is_pre_key_message:
            # If this is not part of a PreKeyMessage, we received a normal Message and can
            # safely delete the OTPK bound to this bare_jid+device.
            self.__state.deleteBoundOTPK(bare_jid, device)
            yield self._storage.storeState(self.__state.serialize())

        # Unpack the message data
        message_data = self.__backend.WireFormat.messageFromWire(message)

        # Get the concatenation of the AES GCM key and tag
        plaintext = session.decryptMessage(
            message_data["ciphertext"],
            message_data["header"]
        )

        # Check the authentication
        self.__backend.WireFormat.finalizeMessageFromWire(
            message,
            {
                "WireFormat": message_data["additional"],
                "DoubleRatchet": plaintext["additional"]
            }
        )

        # Store the changed session
        yield self.__storeSession(bare_jid, device, session)

        plaintext = plaintext["plaintext"]

        aes_gcm_key = plaintext[:16]
        aes_gcm_tag = plaintext[16:]

        promise.returnValue(Cipher(
            algorithms.AES(aes_gcm_key),
            modes.GCM(iv, aes_gcm_tag),
            backend=default_backend()
        ).decryptor())

    ############################
    # public bundle management #
    ############################

    @property
    def public_bundle(self):
        """
        Fill a PublicBundle object with the public bundle data of this State.

        :returns: An instance of ExtendedPublicBundle, filled with the public data of this
            State.
        """

        return self.__state.getPublicBundle()

    @property
    def republish_bundle(self):
        """
        Read, whether this State has changed since it was loaded/since this flag was last
        cleared.

        :returns: A boolean indicating, whether the public bundle data has changed since
            last reading this flag.

        Clears the flag when reading.
        """

        return self.__state.changed

    ######################
    # session management #
    ######################

    @promise.maybe_coroutine(checkSelf)
    def __loadSession(self, bare_jid, device):
        self.__sessions_cache[bare_jid] = self.__sessions_cache.get(bare_jid, {})

        if not (device in self.__sessions_cache[bare_jid]):
            session = yield self._storage.loadSession(bare_jid, device)

            if not session == None:
                session = self.__backend.DoubleRatchet.fromSerialized(session)

            self.__sessions_cache[bare_jid][device] = session

        promise.returnValue(self.__sessions_cache[bare_jid][device])

    @promise.maybe_coroutine(checkSelf)
    def __storeSession(self, bare_jid, device, session):
        self.__sessions_cache[bare_jid] = self.__sessions_cache.get(bare_jid, {})
        self.__sessions_cache[bare_jid][device] = session

        yield self._storage.storeSession(bare_jid, device, session.serialize())

    @promise.maybe_coroutine(checkSelf)
    def __deleteSession(self, bare_jid, device):
        self.__sessions_cache[bare_jid] = self.__sessions_cache.get(bare_jid, {})

        self.__sessions_cache[bare_jid].pop(device, None)

        yield self._storage.deleteSession(bare_jid, device)

    #####################
    # device management #
    #####################

    @promise.maybe_coroutine(checkSelf)
    def __loadActiveDevices(self, bare_jid):
        self.__devices_cache[bare_jid] = self.__devices_cache.get(bare_jid, {})

        if not "active" in self.__devices_cache[bare_jid]:
            devices = yield self._storage.loadActiveDevices(bare_jid)

            self.__devices_cache[bare_jid]["active"] = set(devices)

        promise.returnValue(copy.deepcopy(self.__devices_cache[bare_jid]["active"]))

    @promise.maybe_coroutine(checkSelf)
    def __loadInactiveDevices(self, bare_jid):
        self.__devices_cache[bare_jid] = self.__devices_cache.get(bare_jid, {})

        if not "inactive" in self.__devices_cache[bare_jid]:
            devices = yield self._storage.loadInactiveDevices(bare_jid)
            devices = copy.deepcopy(devices)

            self.__devices_cache[bare_jid]["inactive"] = devices

        promise.returnValue(copy.deepcopy(self.__devices_cache[bare_jid]["inactive"]))

    @promise.maybe_coroutine(checkSelf)
    def __storeActiveDevices(self, bare_jid, devices):
        self.__devices_cache[bare_jid] = self.__devices_cache.get(bare_jid, {})
        self.__devices_cache[bare_jid]["active"] = set(devices)

        yield self._storage.storeActiveDevices(bare_jid, devices)

    @promise.maybe_coroutine(checkSelf)
    def __storeInactiveDevices(self, bare_jid, devices):
        self.__devices_cache[bare_jid] = self.__devices_cache.get(bare_jid, {})
        self.__devices_cache[bare_jid]["inactive"] = copy.deepcopy(devices)
        
        yield self._storage.storeInactiveDevices(bare_jid, devices)

    @promise.maybe_coroutine(checkSelf)
    def newDeviceList(self, bare_jid, active_new):
        active_new = set(active_new)

        if bare_jid == self.__my_bare_jid:
            # The own device can never become inactive
            active_new |= set([ self.__my_device_id ])

        active_old   = yield self.__loadActiveDevices(bare_jid)
        inactive_old = yield self.__loadInactiveDevices(bare_jid)

        devices_old = active_old | set(inactive_old.keys())

        inactive_new = devices_old - active_new

        now = time.time()

        inactive_new = {
            device: inactive_old.get(device, now)
            for device in inactive_new
        }

        yield self.__storeActiveDevices(bare_jid, active_new)
        yield self.__storeInactiveDevices(bare_jid, inactive_new)

        yield self.runInactiveDeviceCleanup()

    @promise.maybe_coroutine(checkSelf)
    def getDevices(self, bare_jid = None):
        yield self.runInactiveDeviceCleanup()

        if bare_jid == None:
            bare_jid = self.__my_bare_jid

        active   = yield self.__loadActiveDevices(bare_jid)
        inactive = yield self.__loadInactiveDevices(bare_jid)

        promise.returnValue({
            "active"   : active,
            "inactive" : inactive
        })

    @promise.maybe_coroutine(checkSelf)
    def __deleteInactiveDevices(self, bare_jid, delete_devices):
        for device in delete_devices:
            yield self.__deleteSession(bare_jid, device)

        inactive_devices = yield self.__loadInactiveDevices(bare_jid)
        
        for device in delete_devices:
            inactive_devices.pop(device, None)

        yield self.__storeInactiveDevices(bare_jid, inactive_devices)

    @promise.maybe_coroutine(checkSelf)
    def deleteInactiveDevicesByQuota(self, per_jid_max = 15, global_max = 0):
        """
        Delete inactive devices by setting a quota. With per_jid_max you can define the
        amount of inactive devices that are kept for each jid, with global_max you can
        define a global maximum for inactive devices. If any of the quotas is reached,
        inactive devices are deleted on an LRU basis. This also deletes the corresponding
        sessions, so if a device comes active again and tries to send you an encrypted
        message you will not be able to decrypt it.

        The value "0" means no limitations/keep all inactive devices.

        It is recommended to always restrict the amount of per-jid inactive devices. If
        storage space limitations don't play a role, it is recommended to not restrict the
        global amount of inactive devices. Otherwise, the global_max can be used to
        control the amount of storage that can be used up by inactive sessions. The
        default of 15 per-jid devices is very permissive, but it is not recommended to
        decrease that number without a good reason.

        This is the recommended way to handle inactive device deletion. For a time-based
        alternative, look at the deleteInactiveDevicesByAge method.
        """

        if per_jid_max < 1 and global_max < 1:
            return

        if per_jid_max < 1:
            per_jid_max = None

        if global_max < 1:
            global_max = None

        bare_jids = yield self._storage.listJIDs()

        if not per_jid_max == None:
            for bare_jid in bare_jids:
                devices = yield self.__loadInactiveDevices(bare_jid)

                if len(devices) > per_jid_max:
                    # This sorts the devices from smaller to bigger timestamp, which means
                    # from old to young.
                    devices = sorted(devices.items(), key = lambda device: device[1])

                    # This gets the first (=oldest) n entries, so that only the
                    # per_jid_max youngest entries are left.
                    devices = devices[:-per_jid_max]

                    # Get the device ids and discard the timestamps.
                    devices = list(map(lambda device: device[0], devices))

                    yield self.__deleteInactiveDevices(bare_jid, devices)
        
        if not global_max == None:
            all_inactive_devices = []

            for bare_jid in bare_jids:
                devices = yield self.__loadInactiveDevices(bare_jid)

                all_inactive_devices.extend(map(
                    lambda device: (bare_jid, device[0], device[1]),
                    devices.items()
                ))

            if len(all_inactive_devices) > global_max:
                # This sorts the devices from smaller to bigger timestamp, which means
                # from old to young.
                devices = sorted(all_inactive_devices, key = lambda device: device[2])

                # This gets the first (=oldest) n entries, so that only the global_max
                # youngest entries are left.
                devices = devices[:-global_max]

                # Get the list of devices to delete for each jid
                delete_devices = {}

                for device in devices:
                    bare_jid  = device[0]
                    device_id = device[1]

                    delete_devices[bare_jid] = delete_devices.get(bare_jid, [])
                    delete_devices[bare_jid].append(device_id)

                # Now, delete the devices
                for bare_jid, devices in delete_devices.items():
                    yield self.__deleteInactiveDevices(bare_jid, devices)

    @promise.maybe_coroutine(checkSelf)
    def deleteInactiveDevicesByAge(self, age_days):
        """
        Delete all inactive devices from the device list storage and cache that are older
        then a given number of days. This also deletes the corresponding sessions, so if
        a device comes active again and tries to send you an encrypted message you will
        not be able to decrypt it. You are not allowed to delete inactive devices that
        were inactive for less than a day. Thus, the minimum value for age_days is 1.

        It is recommended to keep inactive devices for a longer period of time (e.g.
        multiple months), as it reduces the chance for message loss and doesn't require a
        lot of storage.

        The recommended alternative to deleting inactive devices by age is to delete them
        by count/quota. Look at the deleteInactiveDevicesByQuota method for that variant.
        """

        if age_days < 1:
            return

        now = time.time()

        bare_jids = yield self._storage.listJIDs()

        for bare_jid in bare_jids:
            devices = yield self.__loadInactiveDevices(bare_jid)

            delete_devices = []
            for device, timestamp in list(devices.items()):
                elapsed_s = now - timestamp
                elapsed_m = elapsed_s / 60
                elapsed_h = elapsed_m / 60
                elapsed_d = elapsed_h / 24

                if elapsed_d >= age_days:
                    delete_devices.append(device)

            if len(delete_devices) > 0:
                yield self.__deleteInactiveDevices(bare_jid, delete_devices)

    @promise.maybe_coroutine(checkSelf)
    def runInactiveDeviceCleanup(self):
        """
        Runs both the deleteInactiveDevicesByAge and the deleteInactiveDevicesByQuota
        methods with the configuration that was set when calling create.
        """

        yield self.deleteInactiveDevicesByQuota(
            self.__inactive_per_jid_max,
            self.__inactive_global_max
        )

        yield self.deleteInactiveDevicesByAge(self.__inactive_max_age)

    ####################
    # trust management #
    ####################

    @promise.maybe_coroutine(checkSelf)
    def __loadTrust(self, bare_jid, device):
        self.__trust_cache[bare_jid] = self.__trust_cache.get(bare_jid, {})

        if not device in self.__trust_cache[bare_jid]:
            trust = yield self._storage.loadTrust(bare_jid, device)

            self.__trust_cache[bare_jid][device] = None if trust == None else {
                "key"     : base64.b64decode(trust["key"].encode("US-ASCII")),
                "trusted" : trust["trusted"]
            }

        promise.returnValue(copy.deepcopy(self.__trust_cache[bare_jid][device]))

    @promise.maybe_coroutine(checkSelf)
    def __storeTrust(self, bare_jid, device, trust):
        self.__trust_cache[bare_jid] = self.__trust_cache.get(bare_jid, {})
        self.__trust_cache[bare_jid][device] = copy.deepcopy(trust)

        yield self._storage.storeTrust(
            bare_jid,
            device,
            {
                "key"     : base64.b64encode(trust["key"]).decode("US-ASCII"),
                "trusted" : trust["trusted"]
            }
        )

    @promise.maybe_coroutine(checkSelf)
    def __checkTrust(self, bare_jid, device, key):
        trust = yield self.__loadTrust(bare_jid, device)

        if trust == None:
            promise.returnValue(False)

        if not trust["key"] == key:
            promise.returnValue(False)

        promise.returnValue(trust["trusted"])

    @promise.maybe_coroutine(checkSelf)
    def trust(self, bare_jid, device, key):
        yield self.__storeTrust(bare_jid, device, {
            "key"     : key,
            "trusted" : True
        })

    @promise.maybe_coroutine(checkSelf)
    def distrust(self, bare_jid, device, key):
        yield self.__storeTrust(bare_jid, device, {
            "key"     : key,
            "trusted" : False
        })

    #########
    # other #
    #########

    @promise.maybe_coroutine(checkSelf)
    def deleteJID(self, bare_jid):
        """
        Delete all data associated with a JID. This includes the list of active/inactive
        devices, all sessions with that JID and all information about trusted keys.
        """

        yield self.runInactiveDeviceCleanup()

        self.__sessions_cache.pop(bare_jid, None)
        self.__devices_cache.pop(bare_jid, None)
        self.__trust_cache.pop(bare_jid, None)

        yield self._storage.deleteJID(bare_jid)



from __future__ import absolute_import

import asyncio

from .promise import Promise
from .sessionmanager import SessionManager

def wrap(attr):
    def _wrap(*args, **kwargs):
        result = attr(*args, **kwargs)

        if isinstance(result, Promise):
            future = asyncio.Future()
            result.then(future.set_result, future.set_exception)
            return future
        else:
            return result

    return _wrap

class SessionManagerAsyncIO(SessionManager):
    @classmethod
    def create(cls, *args, **kwargs):
        result = super(SessionManagerAsyncIO, cls).create(*args, **kwargs)

        if isinstance(result, Promise):
            future = asyncio.Future()
            result.then(future.set_result, future.set_exception)
            return future
        else:
            return result

    def __getattribute__(self, attr_name):
        attr = super(SessionManagerAsyncIO, self).__getattribute__(attr_name)

        if attr_name in [
            "encryptMessage",
            "encryptKeyTransportMessage",
            "buildSession",
            "decryptMessage",
            "newDeviceList",
            "getDevices",
            "public_bundle",
            "republish_bundle"
        ]:
            return wrap(attr)

        return attr



from __future__ import absolute_import

import x3dh

class X3DHPKEncoder(x3dh.PublicKeyEncoder):
    @staticmethod
    def encodePublicKey(key, key_type):
        """
        Encode given (Montgomery) public key and the type of the key into a sequence of
        bytes.

        :param key: The public key to encode, as a bytes-like object.
        :param key_type: Identification of the curve that this key is used with.
            Currently the only allowed value is (the string) "25519".
        :returns: A bytes-like object, which encodes the public key and possibly its type.
        """

        raise NotImplementedError

    @staticmethod
    def decodePublicKey(key_encoded):
        """
        Decode given (Montgomery) public key and the type of the key from a sequence of
        bytes.

        :param key_encoded: The public key and the type of the key to decode, as a
            bytes-like object.
        :returns: A tuple consisting of a bytes-like object encoding the public key and a
            string identifying the type of the curve this key is used with. Currently the
            only allowed value is (the string) "25519".
        """

        raise NotImplementedError



from __future__ import absolute_import

from .promise import Promise
from .storage import Storage

def makeCallbackPromise(function, *args, **kwargs):
    """
    Take a function that reports its result using a callback and return a Promise that
    listenes for this callback.

    The function must accept a callback as its first parameter.
    The callback must take two arguments:
    - success : True or False
    - result  : The result of the operation if success is True or the error otherwise.
    """

    def _resolver(resolve, reject):
        function(
            lambda success, result: resolve(result) if success else reject(result),
            *args,
            **kwargs
        )

    return Promise(_resolver)

def wrap(is_async, attr):
    def _wrap(*args, **kwargs):
        if is_async:
            return makeCallbackPromise(attr, *args, **kwargs)
        else:
            return attr(None, *args, **kwargs)

    return _wrap

class StorageWrapper(object):
    def __init__(self, wrapped):
        self._wrapped = wrapped

    def __getattribute__(self, attr):
        if attr == "_wrapped":
            return super(StorageWrapper, self).__getattribute__(attr)

        if attr == "is_async":
            return self._wrapped.is_async

        return wrap(self.is_async, getattr(self._wrapped, attr))



class Storage(object):
    """
    The interface used by the SessionManager to persist data between runs.

    There are two possible ways to implement the Storage class: synchronous or
    asynchronous.

    The mode is determined by the result of the is_async method.

    If the implementation is synchronous, the callback parameter is None.

    If the implementation is asynchronous, the callback parameter is a function that takes
    two arguments:
    - success: True or False
    - result: The result of the operation if success is True or the error if success is
        False

    Note:
    The SessionManager does caching to reduce the number of calls to a minimum. There
    should be no need to add caching or any other logic in here, just plain storing and
    loading.
    """

    def loadOwnData(self, callback):
        """
        Load the own data.

        Return a dictionary of following structure:
        {
            "own_bare_jid"  : string,
            "own_device_id" : int
        }

        or None, if no own data was stored previously.
        """

        raise NotImplementedError

    def storeOwnData(self, callback, own_bare_jid, own_device_id):
        """
        Store given own data, overwriting previously stored data.
        """

        raise NotImplementedError

    def loadState(self, callback):
        """
        Load the state.

        Return the stored structure or None, if no state was stored previously.
        """

        raise NotImplementedError

    def storeState(self, callback, state):
        """
        Store the state, overwriting the old state, if it exists.

        state is passed as a serializable object, that means it consist of a combination
        of the following types:
        - dictionaries
        - lists
        - strings
        - integers
        - floats
        - booleans
        - None

        You can dump this object using for example the json module.

        For more information on how the state object is structured, look at the
        omemo.X3DHDoubleRatchet.serialize method.
        """

        raise NotImplementedError

    def loadSession(self, callback, bare_jid, device_id):
        """
        Load a session with given bare_jid and device id.

        bare_jid is passed as a string, device_id as an integer.

        Return either the structure previously stored or None.
        """

        raise NotImplementedError

    def storeSession(self, callback, bare_jid, device_id, session):
        """
        Store a session for given bare_jid and device id, overwriting the previous
        session, if it exists.

        bare_jid is passed as string, device_id as an integer.

        session is passed as a serializable object, that means it consist of a combination
        of the following types:
        - dictionaries
        - lists
        - strings
        - integers
        - floats
        - booleans
        - None

        You can dump this object using for example the json module.
        """

        raise NotImplementedError

    def deleteSession(self, callback, bare_jid, device_id):
        """
        Completely wipe the session associated with given bare_jid and device_id from the
        storage.

        bare_jid is passed as string, device_id as an integer.
        """

        raise NotImplementedError

    def loadActiveDevices(self, callback, bare_jid):
        """
        Load the list of active devices for a given bare_jid.

        An "active device" is a device, which is listed in the most recent version of
        the device list pep node.

        bare_jid is passed as a string, the result is expected to be a list of integers.
        """

        raise NotImplementedError

    def loadInactiveDevices(self, callback, bare_jid):
        """
        Load the list of inactive devices for a given bare_jid.

        An "inactive device" is a device, which was listed in an older version of
        the device list pep node, but is NOT listed in the most recent version.

        bare_jid is passed as a string, the result is expected to be a dict mapping from
        int to int, where the keys are device ids and the values are timestamps (seconds
        since epoch).
        """

        raise NotImplementedError

    def storeActiveDevices(self, callback, bare_jid, devices):
        """
        Store the active devices for given bare_jid, overwriting the old stored list,
        if it exists.

        bare_jid is passed as a string, devices as a list of integers.
        """

        raise NotImplementedError

    def storeInactiveDevices(self, callback, bare_jid, devices):
        """
        Store the inactive devices for given bare_jid, overwriting the old stored list,
        if it exists.

        bare_jid is passed as a string, devices as a dict mapping from int to int, where
        the keys are device ids and the values are timestamps (seconds since epoch).
        """

        raise NotImplementedError

    def storeTrust(self, callback, bare_jid, device_id, trust):
        """
        bare_jid: string
        device_id: int

        trust: {
            "key"     : string (Base64 encoded bytes),
            "trusted" : bool
        }
        """

        raise NotImplementedError

    def loadTrust(self, callback, bare_jid, device_id):
        """
        """

        raise NotImplementedError

    def listJIDs(self, callback):
        """
        List all bare jids that have associated device lists stored in the storage.
        It doesn't matter if the lists are empty or not.

        Return a list of strings.
        """

        raise NotImplementedError

    def deleteJID(self, callback, bare_jid):
        """
        Delete all data associated with given bare_jid. This includes the active and
        inactive devices, all sessions stored for that jid and all information about
        trusted keys.
        """

        raise NotImplementedError

    @property
    def is_async(self):
        """
        Return, whether this implementation is asynchronous.

        Read the introduction to this module above for details on what this value changes.
        """

        raise NotImplementedError



class NoEligibleDevicesException(Exception):
    def __init__(self, bare_jid):
        self.__bare_jid = bare_jid

    @property
    def bare_jid(self):
        return self.__bare_jid

    def __eq__(self, other):
        return (
            isinstance(other, NoEligibleDevicesException) and
            other.bare_jid == self.bare_jid
        )

    def __hash__(self):
        return hash(self.bare_jid)



class MissingBundleException(Exception):
    def __init__(self, bare_jid, device):
        self.__bare_jid = bare_jid
        self.__device   = device

    @property
    def bare_jid(self):
        return self.__bare_jid

    @property
    def device(self):
        return self.__device

    def __eq__(self, other):
        return (
            isinstance(other, MissingBundleException) and
            other.bare_jid == self.bare_jid and
            other.device == self.device
        )

    def __hash__(self):
        return hash((self.bare_jid, self.device))



from __future__ import absolute_import

from .exceptions import UnknownKeyException

class ExtendedPublicBundle(object):
    """
    This class looks exactly the same as the PublicBundle class, but the types of the
    fields are a bit different:

    The spk field is not a key, but a dictionary containing the key and the id:
    spk = {
        "key" : key,
        "id"  : id
    }

    The otpks field is not an array of keys, but an array of dictionaries containing the
    key and the id:
    otpks = [
        {
            "key" : key,
            "id"  : id
        },
        {
            "key" : key,
            "id"  : id
        },
        ...
    ]
    """

    def __init__(self, ik, spk, spk_signature, otpks):
        self.__ik = ik
        self.__spk = spk
        self.__spk_signature = spk_signature
        self.__otpks = otpks

    @classmethod
    def parse(cls, backend, ik, spk, spk_signature, otpks):
        """
        Use this method when creating a bundle from data you retrieved directly from some
        PEP node. This method applies an additional decoding step to the public keys in
        the bundle. Pass the same structure as the constructor expects.
        """

        ik = backend.decodePublicKey(ik)[0]

        spk["key"] = backend.decodePublicKey(spk["key"])[0]

        otpks = list(map(lambda otpk: {
            "key" : backend.decodePublicKey(otpk["key"])[0],
            "id"  : otpk["id"]
        }, otpks))

        return cls(ik, spk, spk_signature, otpks)

    def serialize(self, backend):
        """
        Use this method to prepare the data to be uploaded directly to some PEP node. This
        method applies an additional encoding step to the public keys in the bundle. The
        result is a dictionary with the keys ik, spk, spk_signature and otpks. The values
        are structured the same way as the inputs of the constructor.
        """

        return {
            "ik": backend.encodePublicKey(self.ik, "25519"),
            "spk": {
                "id"  : self.spk["id"],
                "key" : backend.encodePublicKey(self.spk["key"], "25519"),
            },
            "spk_signature": self.spk_signature,
            "otpks": list(map(lambda otpk: {
                "id"  : otpk["id"],
                "key" : backend.encodePublicKey(otpk["key"], "25519")
            }, self.otpks))
        }

    @property
    def ik(self):
        return self.__ik

    @property
    def spk(self):
        return self.__spk

    @property
    def spk_signature(self):
        return self.__spk_signature

    @property
    def otpks(self):
        return self.__otpks

    def findOTPKId(self, otpk):
        otpks = list(filter(lambda x: x["key"] == otpk, self.otpks))

        if len(otpks) != 1:
            raise UnknownKeyException("Tried to get the id of an unknown OTPK.")

        return otpks[0]["id"]

    def findSPKId(self, spk):
        # If the requested spk is the one contained in this bundle...
        if self.spk["key"] == spk:
            # ...return the id
            return self.spk["id"]

        raise UnknownKeyException("Tried to get the id of an unknown SPK.")

    def __eq__(self, other):
        try:
            return (
                self.ik == other.ik and
                self.spk == other.spk and
                self.spk_signature == other.spk_signature and
                self.otpks == other.otpks
            )
        except:
            return False
